
/**
 * Copyright (C) 2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 9 (sqlite cache)
 */

#include "squidwall.h"
#include "timestamp.h"

#if WANT_SQLITE

#include "sqlite3.h"

static pthread_mutex_t db_mutex = PTHREAD_MUTEX_INITIALIZER;
static sqlite3 *db;

static pthread_mutex_t errmsg_mutex = PTHREAD_MUTEX_INITIALIZER;
static char *errmsg;

static pthread_mutex_t status_mutex = PTHREAD_MUTEX_INITIALIZER;
static int status=0;

#define STATUS_DB_OPEN   1
#define STATUS_DB_CLOSED 2
#define STATUS_DB_BUSY   4

void sql_init(void)
{
	stralloc sa;
	int rv;

	if (isset(status, STATUS_DB_OPEN)) {
		return;
	}

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, CACHE_DIR)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, "cache.db")) die_nomem();
	sa.s[sa.len]=0;

	pthread_mutex_lock(&db_mutex);
	rv = sqlite3_open(sa.s, &db);
	pthread_mutex_unlock(&db_mutex);
	stralloc_free(&sa);

	if (rv == SQLITE_OK) {
		pthread_mutex_lock(&status_mutex);
		set(status, STATUS_DB_OPEN);
		pthread_mutex_unlock(&status_mutex);
		debug(9,6, "sql_init(): OPEN");
	}

	return;
}

int sql_exec(stralloc *sql)
{
	int rv;

	if (!isset(status, STATUS_DB_OPEN)) {
		return -1;
	}

	/* ensure termination */
	sql->s[sql->len]=0;

	debug(9,6, "sql_exec(), todo: [%s]", sql->s);
	again:
	pthread_mutex_lock(&db_mutex);
	rv = sqlite3_exec(db, sql->s, 0, 0, &errmsg);
	pthread_mutex_unlock(&db_mutex);
	
	switch (rv) {
	case SQLITE_OK:
		break;
	case SQLITE_BUSY:
		pthread_mutex_lock(&errmsg_mutex);
		debug(9,6, "sql_exec(), busy: [%s]", errmsg);
		sqlite3_free(errmsg);
		pthread_mutex_lock(&errmsg_mutex);
		sqlite3_sleep(sqlite_timeout);
		goto again;
	default:
		pthread_mutex_lock(&errmsg_mutex);
		debug(9,6, "sql_exec(), error: [%s]", errmsg);
		sqlite3_free(errmsg);
		pthread_mutex_unlock(&errmsg_mutex);
		return -1;
	}

	return 0;
}

void sql_quit(void)
{
	int rv = STATUS_DB_BUSY;

	for (;;) {
		pthread_mutex_lock(&db_mutex);
		rv = sqlite3_close(db);
		pthread_mutex_unlock(&db_mutex);
		switch (rv) {
		case SQLITE_OK:
		case SQLITE_ERROR:
			pthread_mutex_lock(&status_mutex);
			status=STATUS_DB_CLOSED;
			pthread_mutex_unlock(&status_mutex);
			debug(9,6, "sql_quit(): CLOSED");
			return;
		case SQLITE_BUSY:
			pthread_mutex_lock(&status_mutex);
			set(status, STATUS_DB_BUSY);
			pthread_mutex_unlock(&status_mutex);
			debug(9,6, "sql_quit(): BUSY");
		}
	}

	return;
}

#endif
