
/**
 * Copyright (C) 2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 7
 */

#include "squidwall.h"
#include "hextab.h"
#include "md5.h"

/**
 * The cache is implemented as a static array with N members.
 * - the cache isn't ordered in any manner
 * - we do a linear search for some item
 *   - on this search, we remember the oldest entry and delete it later (if the
 *     cache size is reached)
 */
static pthread_mutex_t array_memcache_mutex = PTHREAD_MUTEX_INITIALIZER;
static array array_memcache;

/**
 * memcache_get_hash - generate hash value of the squid variables
 *
 * @param r ...
 * @param hash the null-terminated md5 hash of the variables
 */
static void memcache_get_hash(request_t *r, char *hash)
{
	unsigned char digest[16];
	MD5_CTX md5;
	char *e;
	int i;

	MD5_Init(&md5);
	MD5_Update(&md5, r->squid_url,    str_len(r->squid_url));
	MD5_Update(&md5, r->squid_ip,     str_len(r->squid_ip));
	MD5_Update(&md5, r->squid_host,   str_len(r->squid_host));
	MD5_Update(&md5, r->squid_user,   str_len(r->squid_user));
	MD5_Update(&md5, r->squid_method, str_len(r->squid_method));
	MD5_Final(digest, &md5);

	e=hash;
	for (i=0; i<16; i++) {
		*e = hextab[digest[i]/16]; ++e;
		*e = hextab[digest[i]%16]; ++e;
	}
	*e=0;

	return;
}


/**
 * memcache_init - allocates memory for the cache
 */
void memcache_init(request_t *r)
{
	struct memcache *mc;
	int i;

	if (!memcache) return;
	debug(7,5, "memcache_init()");

	pthread_mutex_lock(&array_memcache_mutex);
	mc=array_allocate(&array_memcache, sizeof(struct memcache),
			memcache*sizeof(struct memcache));
	if (!mc) die_nomem();
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);
		stralloc_init(&mc->url);
		mc->ts=0;
	}
	pthread_mutex_unlock(&array_memcache_mutex);

	return;
}


/**
 * memcache_insert - insert some new entry in our small memory cache
 *
 * @param result - gets filled with the url to be redirected to
 * @return: MEMCACHE_MISS there was no such entry in the cache
 *          MEMCACHE_HIT  the current url was found, result is set
 */
void memcache_insert(request_t *r)
{
	struct memcache *mc=0;
	time_t oldest=0;
	int i, id_old=0;
	char hash[33];

	if (!memcache) return;
	debug(7,5, "memcache_insert()");

	memcache_get_hash(r, hash);
	pthread_mutex_lock(&array_memcache_mutex);
	mc = array_get(&array_memcache, sizeof(struct memcache), 0);
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);

		if (mc->ts == 0) break; /* last entry */

		/* check, if this is the oldest entry */
		if (oldest) {
			if (mc->ts > oldest) oldest = mc->ts;
		} else {
			oldest = mc->ts;
			id_old=i;
		}

		if (str_equal(mc->hash, hash)) {
			debug(7,4, "memcache_insert() REFRESH OLD [%s] [%s]", hash, r->squid_url);
			if (!stralloc_copy(&mc->url, &r->redirsite)) die_nomem();
			mc->ts=time(0);
			return;
		}
	}

	/* add new entry @end of list */
	if (mc->ts == 0) {
		debug(7,4, "memcache_insert() ADD@END [%s] [%s]", hash, r->squid_url);
		if (!stralloc_copy(&mc->url, &r->redirsite)) die_nomem();
		mc->hash[fmt_str(mc->hash, hash)]=0;
		mc->ts=time(0);
	} else {
		mc = array_get(&array_memcache, sizeof(struct memcache), id_old);
		debug(7,4, "memcache_insert() REPLACE OLD [%s] [%s]", hash, r->squid_url);
		if (!stralloc_copy(&mc->url, &r->redirsite)) die_nomem();
		mc->hash[fmt_str(mc->hash, hash)]=0;
		mc->ts=time(0);
	}
	pthread_mutex_unlock(&array_memcache_mutex);

	return;
}


/**
 * memcache_reinit - free old memory cache and create a new one
 */
void memcache_reinit(request_t *r)
{
	struct memcache *mc;
	int i;

	if (!memcache) return;
	debug(7,5, "memcache_reinit()");

	pthread_mutex_lock(&array_memcache_mutex);
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);
		stralloc_free(&mc->url);
	}

	if (array_bytes(&array_memcache))
		array_reset(&array_memcache);
	pthread_mutex_unlock(&array_memcache_mutex);

	memcache_init(r);

	return;
}


/**
 * memcache_lookup - look up some entry in our small memory cache
 *
 * @param result gets filled with the url to be redirected to
 * @return: MEMCACHE_MISS there was no such entry in the cache
 *          MEMCACHE_HIT  the current url was found, result is set
 */
int memcache_lookup(request_t *r)
{
	struct memcache *mc;
	char hash[33];
	int i;

	if (!memcache) return MEMCACHE_MISS;

	memcache_get_hash(r, hash);

	pthread_mutex_lock(&array_memcache_mutex);
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);
		if (mc->ts == 0) break; /* last entry */
		if (str_equal(mc->hash, hash)) {
			debug(7,5, "MEMCACHE_HIT: %s\n", r->squid_url);
			if (!stralloc_cat(&r->redirsite, &mc->url)) die_nomem();
			mc->ts=time(0);
			return MEMCACHE_HIT;
		}
	}
	pthread_mutex_unlock(&array_memcache_mutex);

	debug(7,5, "MEMCACHE_MISS");

	return MEMCACHE_MISS;
}
