#ifndef SW_SQUIDWALL_H
#define SW_SQUIDWALL_H

/**
 * Copyright (C) 2005-2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "params.h"

#define VERSION    "squidwall 0.5a"

/* system */
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/un.h>

#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <dirent.h>
#include <sched.h>
#include <regex.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

/* libowfat */
#include "array.h"
#include "buffer.h"
#include "case.h"
#include "fmt.h"
#include "ip4.h"
#include "ip6.h"
#include "io.h"
#include "str.h"
#include "open.h"
#include "scan.h"
#include "socket.h"
#include "stralloc.h"
#include "textcode.h"
#include "cache.h"

/* squidwall */
#include "request.h"
#include "sig.h"
#include "log.h"
#include "env.h"
#include "list.h"
#include "control.h"
#include "memcache.h"
#include "strallocX.h"
#include "thread.h"
#include "check.h"
#include "read_acls.h"
#include "read_controls.h"

/* type of acl directories */
#define TYPE_ACL      0
#define TYPE_GLOBAL   1
#define TYPE_STRING   2
#define TYPE_REGEX    3
#define TYPE_PHRASE   4
#define TYPE_REDIRECT 5

/* hosts/ip's + users */
struct squidwall_acl {
	struct list_head list;     /* the list with all elements */
	stralloc filename;         /* without the '?' prefix */

	int flags;
#define ACL_NONE      0            /* no flag is set currently */
#define ACL_ALLOW     1            /* the host/user isn't filtered @all */
#define ACL_FILTER    2            /* the host/user has filtered access */
#define ACL_DENY      4            /* the host/user can't use the proxy */
#define ACL_REDIR     8            /* redirect with '302:' prefix (ssl and stuff) */
#define ACL_MULTI    16            /* redirect with '#nr ...' - new redirector interface */
#define ACL_MEMHIT   32            /* doesn't need memcache_insert() */

	int todo;                  /* TODO_(NONE|VIRUS|CONT...) */
#define TODO_NONE        0
#define TODO_PHRASE      1
#define TODO_CONT        2
#define TODO_CONTr       4
#define TODO_URL         8
#define TODO_URLr       16
#define TODO_URLmd5     32
#define TODO_URLsha1    64
#define TODO_BAN       128
#define TODO_BANr      256
#define TODO_BANc      512
#define TODO_WLST     1024
#define TODO_WLSTr    2048
#define TODO_VIRUS    4096

/* per default, all files of all filters are used! */
#define TODO_DEFAULT (1|2|4|8|16|32|64|128|256|512|1024|2048|4096)

	/* filter which redirect to a denypage */
	stralloc pfilter;      /* +p / -p */
	stralloc cfilter;      /* +c / -c */
	stralloc crfilter;     /* +C / -C */
	stralloc ufilter;      /* +u / -u */
	stralloc ufilter_md5;  /* +m / -m */
	stralloc ufilter_sha1; /* +s / -s */
	stralloc urfilter;     /* +U / -U */

	/* filter which redirect to an empty image file */
	stralloc bfilter;      /* +b / -b */
	stralloc bcfilter;     /* +a / -a */
	stralloc brfilter;     /* +B / -B */

	/* whitelists */
	stralloc wfilter;      /* +w / -w */
	stralloc wrfilter;     /* +W / -W */
};

/* bfilter, ufilter, cfilter, ufilter_md5, ufilter_sha1, wfilter  */
struct squidwall_string {
	struct list_head list;  /* the list with all elements */
	stralloc filename;      /* just the filename of the filters */
	stralloc strings;       /* all strings from filename */
};

/* pfilter */
struct squidwall_phrase {
	struct list_head list;  /* the list with all elements */
	stralloc filename;      /* just the filename of the phrase filters */
	stralloc strings;       /* all string:hitpoint pairs from filename */
	array points;           /* array of type 'unsigned int' (0..255) */
};

/* bcfilter, brfilter, urfilter, crfilter, wrfilter */
struct squidwall_regex {
	struct list_head list;  /* the list with all elements */
	stralloc filename;      /* just the filename of the regex filters */
	array regex;            /* array of type 'struct named_regex' */
};

struct named_regex {
	stralloc n;  /* the string, which defines the regex */
	regex_t r;   /* the precompiled regex */
	int cflags;
	int eflags;
};

/* lists with all defined filters */
extern struct list_head hosts;
extern struct list_head users;
extern struct list_head redirects;

extern struct list_head bfilter;
extern struct list_head bcfilter;
extern struct list_head brfilter;
extern struct list_head cfilter;
extern struct list_head crfilter;
extern struct list_head pfilter;
extern struct list_head ufilter;
extern struct list_head ufilter_md5;
extern struct list_head ufilter_sha1;
extern struct list_head urfilter;
extern struct list_head wfilter;
extern struct list_head wrfilter;

/* case_strstr.c -> 1=>found | 0=>not found */
extern int case_strstr(const char *haystack, const char *needle);

/* redir_site.c */
#define REDIR_TYPE_USER    0 /* user not okay */
#define REDIR_TYPE_HOST    1 /* host not okay */
#define REDIR_TYPE_URL     2 /* url not okay */
#define REDIR_TYPE_CONT    4 /* content not okay */
#define REDIR_TYPE_VIRUS   8 /* virus found */
#define REDIR_TYPE_BANNER 16 /* banner found */
#define REDIR_TYPE_CONFIG 32 /* webinterface */
void redir_site(request_t *r, int type);

int setup_acls(request_t *r, int user, int host, int global);

#define isset(v,x)     (((v)&(x)) == (x))
#define set(v,x)       (v) |=  (x)
#define unset(v,x)     (v) &= ~(x)

extern int read_controls(char *path, time_t *mtime);

extern void read_filters_file(struct squidwall_acl *acl,
				const char *path, const char *file);

extern void read_filters_acl(struct squidwall_acl *acl,
				const char *path, const char *file);

/* dummy request for debugging / logging from the main thread */
extern request_t *r;

#endif
