
/**
 * Copyright (C) 2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 6
 */

#include "squidwall.h"
#include "thread.h"
#include "request.h"

/**
 * this structure is for passing work to a squidwall thread
 */
struct squidwall_work {
	struct list_head list;    /* when idle, on squidwall_idle_list */
	void (*fn)(request_t *r); /* callback function */
	int nr;                   /* worker number */
};

/**
 * thread management
 *
 * - the minimum number of squidwall threads is defined by NUMTHREADS (params.h)
 * - if needed, we just create new threds
 * - we never destroy threads, so we don't have to restart them later
 */
static void thread_start_one_worker(void);
static void *thread_worker(void *r);

/**
 * havework contition and its mutex
 */
static pthread_cond_t havework_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t havework_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * current number of squidwall threads and its mutex
 */
static int nr_squidwall_threads = 0;
static pthread_mutex_t nr_squidwall_threads_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * this list maintains the idle threads, protected with a mutex
 */
static LIST_HEAD(squidwall_idle_list);
static pthread_mutex_t squidwall_idle_list_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * pointer for passing the work to some worker, also with a mutex
 */
static struct squidwall_work *work;
static pthread_mutex_t work_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * thread_init - initialize the thread pool
 */
void thread_init(void)
{
	int i;

	/* start our minimum of threads */
	for (i=0; i<NUMTHREADS; i++)
		thread_start_one_worker();

	return;
}

/**
 * thread_operation - wake up a squidwall thread and put some work to it
 *
 * @param fn function to execute
 * @param arg argument to the function
 * @return
 *   1 = we started some new worker, but need a recall
 *   0 = everything was okay
 *  -1 = some error detected
 */
int thread_operation(void (*fn)(request_t *r))
{
	struct squidwall_work *w;

	/* no function given */
	if (fn == NULL) return -1;

	pthread_mutex_lock(&squidwall_idle_list_mutex);

	if (list_empty(&squidwall_idle_list)) {
		struct timespec t;

		debug(6,6,"starting a new worker!");
		thread_start_one_worker();
		t.tv_sec=0; t.tv_nsec=30;
		nanosleep(&t,&t); /* let the thread start */

		/* we fall through and know, that there _is_ an new worker! */
	}
    
    	w = list_entry(squidwall_idle_list.next, struct squidwall_work, list);
	list_del_init(&w->list); /* remove from idle list */

	
	/* if the list is now empty, create a new one, just to have one later ;) */
	if (list_empty(&squidwall_idle_list)) {
		debug(6,6,"starting a new _extra_ thread!");
		thread_start_one_worker();
	}
	pthread_mutex_unlock(&squidwall_idle_list_mutex);

	debug(6,6,"worker(%d) was idle...", w->nr);
	w->fn = fn;

	debug(6,6,"worker(%d) gets new work...", w->nr);
	pthread_mutex_lock(&work_mutex);
	work = w;
	pthread_cond_signal(&havework_cond); /* give the work to some worker */

	return 0;
}


/**
 * thread_worker - maintain one thread in an endless loop
 *
 * @param nr  the number of currently running threads
 * @return    never
 */
static void *thread_worker(void *nr)
{
	struct squidwall_work *w;
	request_t *r;
	int rv = 1;

	r = (request_t *)malloc(sizeof(request_t));
	w = (struct squidwall_work *)malloc(sizeof(struct squidwall_work));
	log_init(r,0);

	INIT_LIST_HEAD(&w->list);
	w->nr = (int)nr;
	w->fn = NULL;

	debug(6,3,"worker(%d) joins the threadpool...", w->nr);

	for (;;) {
		pthread_mutex_lock(&squidwall_idle_list_mutex);
		debug(6,5,"worker(%d) is now on idle list...", w->nr);
		list_move(&w->list, &squidwall_idle_list); /* add to idle list */
		pthread_mutex_unlock(&squidwall_idle_list_mutex);

		/* wait for work */
		pthread_mutex_lock(&havework_mutex);
		sched_yield();
		rv = pthread_cond_wait(&havework_cond, &havework_mutex);
		debug(6,5,"worker(%d) got some work...", w->nr);

		if (rv) {
			debug(6,1,"worker(%d) Oops! errno=%d", w->nr, errno);
			die_pthread();
		}

		/* we got now some work */
		w = work;
		pthread_mutex_unlock(&work_mutex);

		if (w->fn == NULL) {
			debug(6,5,"worker(%d), NULL work function!", w->nr);
			die_pthread();
		}

		/* execute some function */
		debug(6,5,"worker(%d) handles request", w->nr);
		(*w->fn)(r);
		debug(6,5,"worker(%d) has finished", w->nr);

		pthread_mutex_unlock(&havework_mutex);
	}

	/* never reached */
	die_pthread();
}

static void thread_start_one_worker(void)
{
	pthread_t p;
	int rv;

	pthread_mutex_lock(&nr_squidwall_threads_mutex);
	nr_squidwall_threads++;
	pthread_mutex_unlock(&nr_squidwall_threads_mutex);

	/* creates a new thread, which exits only on program termination! */
	rv = pthread_create(&p, 0, thread_worker, (void*)nr_squidwall_threads);
	if (rv) die_pthread();

	return;
}
